From iris.algebra Require Import gmap frac deprecated auth gset.
From iris.base_logic Require Import big_op wsat invariants.
From iris.heap_lang Require Import proofmode notation.
From fletch Require Import definitions.
Import uPred.
Import dec_agree.

Section monoids.
  
(* ABSTRACT STATE 
===================================================*)

  (* The process queue RA link *)
  Definition linkR : cmraT := authR (gmapUR loc (dec_agreeR gname)).

  (* The process in queue token CMRA - the gname above refers to instantiations
   of this RA. *)
  Definition inqueueR : cmraT := authR (gset_disjUR procAddrT).

  (* The list token CMRA *)
  Definition listR : cmraT := (prodR fracR (dec_agreeR (list procAddrT))).

  (* The process CMRA *) 
  Definition procR : cmraT :=
    authR (gmapUR procAddrT (prodR fracR (dec_agreeR procValT))).

(* Use one combined class *)
  Class fletchG Σ := FletchG {
    link_tokG :> inG Σ linkR;
    inqueue_tokG :> inG Σ inqueueR;
    list_tokG :> inG Σ listR;
    proc_tokG :> inG Σ procR
  }.

  Definition LinkAll qs : linkR :=
    ● qs.

  Definition Link (l : loc) (γ : gname) : linkR :=
    ◯ {[ l := (DecAgree γ) ]}.

  Definition InQueueAuth (l : list procAddrT) : inqueueR := 
    ● GSet (of_list' l) .

  Definition InQueue (p : procAddrT) : inqueueR :=
    ◯ GSet {[ p ]}.

  Definition List (l : list procAddrT) : listR := 
    (½, DecAgree l).

  Definition ProcAuth h : procR :=
    ● h.
    
  Definition Proc (x : procAddrT) (π : Qp) (v : procValT) : procR :=
    ◯ {[ x := (π, DecAgree v) ]}.

  (* Context {Σ : gFunctors} `{!fletchG Σ} (N : namespace) (γp γq γl : gname). *)
  (* Context `{heapN ⊥ N}. *)

  (* Context {Σ : gFunctors} `{!fletchG Σ} (γp γq γl : gname). *)

  Context {Σ : gFunctors} `{!fletchG Σ}.

  Local Notation iProp := (iProp Σ).

  Global Instance TimelessP_match X (p : option X) (a : iProp) (b : X → iProp) :
    TimelessP a → (∀ x, TimelessP (b x)) →
    TimelessP (match p with | None => a | Some x => b x end).
  Proof.
    intros ? ? ; destruct p ; auto.
  Qed.


(* GHOST STATE HELPERS 
===================================================*)

  Lemma inQueue_combine γ h (p : procAddrT) :
    own γ (InQueueAuth h) ∗
    own γ (InQueue p)
    ⊢ ⌜ (p ∈ h) ⌝. 
  Proof.
    rewrite /InQueueAuth /InQueue.
    iIntros "[H1 H2]". 
    iDestruct (own_valid_2 with "H1 H2") as %Hvalid.
    apply auth_valid_discrete_2 in Hvalid as [Hx Hx'].
    apply gset_disj_included in Hx.
    iPureIntro.
    by apply elem_of_subseteq_singleton, elem_of_of_list in Hx.
  Qed.
  
  Lemma inQueue_add l p γ :
    p ∉ l → own γ (InQueueAuth l)
    ⊢ |==> own γ (InQueueAuth (l++[p])) ∗ own γ (InQueue p).
  Proof.
    intros; iIntros "H".
    rewrite /InQueueAuth /InQueue -own_op -of_list'_union_tail. 
    iDestruct (own_update γ
                          (● GSet (of_list' l))
                          (● GSet ({[p]} ∪ of_list' l) ⋅ ◯ GSet {[p]}) with "H") as "$".
    apply auth_update_alloc.
      by apply gset_disj_alloc_empty_local_update, disjoint_singleton_l,
         not_elem_of_of_list.
  Qed.

  Lemma inQueue_remove (p : procAddrT) (l1 l2 : list procAddrT) γ :
    {[p]} ⊥ (of_list' (l1++l2)) →
    own γ (InQueueAuth (l1 ++ p::l2)) ∗
    own γ (InQueue p)
    ⊢ |==> own γ (InQueueAuth (l1++l2)).
  Proof.
    intros; iIntros "H".
    rewrite /InQueueAuth /InQueue -own_op of_list'_middle.
    iDestruct (own_update γ
                          (● GSet ({[p]} ∪ of_list (l1++l2)) ⋅ ◯ GSet {[p]})
                          (● GSet (of_list (l1++l2)))) as "HX";
      last by iDestruct ("HX" with "H") as "H".
    rewrite -gset_disj_union.
    apply auth_update_dealloc, gset_disj_dealloc_empty_local_update.
    apply H.
  Qed.

  Lemma inQueue_remove_cons (p : procAddrT) (l : list procAddrT) γ :
    {[p]} ⊥ (of_list' l) →
    own γ (InQueueAuth (p::l)) ∗
    own γ (InQueue p)
    ⊢ |==> own γ (InQueueAuth l).
  Proof.
    apply (inQueue_remove p [] l).
  Qed.

  Lemma link_split l γq γ :
    own γq (Link l γ) ⊢
    own γq (Link l γ) ∗ own γq (Link l γ).
  Proof.
    iIntros "H".
      by rewrite /Link -own_op -auth_frag_op op_singleton dec_agree_idemp.
  Qed.

  Lemma link_eq l γq γ γ' :
    own γq (Link l γ) ∗ own γq (Link l γ') ⊢ ⌜ γ = γ' ⌝.
  Proof.
    iIntros "(H1 & H2)".
    iDestruct (own_valid_2 with "H1 H2") as %foo.
    rewrite /Link -auth_frag_op op_singleton in foo.
    specialize (foo l). rewrite /= lookup_singleton in foo.
    apply dec_agree_op_inv in foo.
      by inversion foo.
  Qed.
  
  Lemma list_alloc l :
    True ⊢ |==> ∃ γ, own γ (List l) ∗ own γ (List l).
  Proof.
    iIntros.
    iMod (own_alloc ((List l) ⋅ (List l))) as (γ) "Hown".
      by rewrite /List pair_op dec_agree_idemp.
      rewrite own_op.
      iModIntro. by iExists _.
  Qed.
  
  Lemma list_combine γ l l' :
    own γ (List l) ∗ own γ (List l') ⊢ ⌜ l = l' ⌝.
  Proof.
    iIntros "H". rewrite -own_op /List pair_op.
    iDestruct (own_valid with "H") as %[H1 H2]. simpl in *.
    apply dec_agree_op_inv in H2. by inversion H2.
  Qed.

  Lemma list_update γ l l' :
    own γ (List l) ∗ own γ (List l) ⊢
        |==> own γ (List l') ∗ own γ (List l').
  Proof.
    iIntros "H".
    rewrite -own_op /List pair_op frac_op' dec_agree_idemp Qp_div_2.
    iDestruct (own_valid with "H") as %Hvalid.
    rewrite -own_op /List pair_op frac_op' dec_agree_idemp Qp_div_2.
    iApply (own_update with "H").
    eapply cmra_update_exclusive. done. 
  Qed.
  
  Lemma proc_own_combine γp p π π' v v' : 
    own γp (Proc p π v) ∗ own γp (Proc p π' v') ⊢ ⌜v = v'⌝ ∧ own γp (Proc p (π + π') v).
  Proof.
    iIntros "H".
    rewrite -own_op /Proc -auth_frag_op op_singleton pair_op.
    iDestruct (own_valid with "H") as %Hvalid.
    apply singleton_valid in Hvalid as [_ Hvalid2]; simpl in *.
    apply dec_agree_op_inv in Hvalid2 as [= ->].
    rewrite dec_agree_idemp.
    auto.
  Qed.
  
  Lemma proc_own_full γp p v v' :
    own γp (Proc p ¼ v) ∗ own γp (Proc p ¾ v') ⊢ ⌜v = v'⌝ ∧ own γp (Proc p 1 v).
  Proof.
    iIntros "H".
    iDestruct (proc_own_combine with "H") as "H". simpl.
    rewrite Qp_div_S Qp_plus_comm -Qp_plus_assoc Qp_div_S Qp_div_2. auto.
  Qed.

  Lemma proc_own_split γp p v :
    own γp (Proc p 1 v) ⊢ own γp (Proc p ¼ v) ∗ own γp (Proc p ¾ v).
  Proof.
    iIntros "H".
    rewrite -own_op /Proc -auth_frag_op op_singleton pair_op.
    replace (¼ ⋅ ¾) with (1)%Qp.
    rewrite dec_agree_idemp. auto.
      by rewrite Qp_div_S Qp_plus_comm /op /cmra_op /= /frac_op Qp_plus_assoc
                 Qp_div_S Qp_div_2.
  Qed.

  Lemma proc_own_eq γp p v v' :
    own γp (Proc p ¼ v) ∗ own γp (Proc p ¾ v')
        ⊢ ⌜v = v'⌝.
  Proof.
    iIntros "H". by iDestruct (proc_own_full with "H") as "[% H2]". 
  Qed.

  Lemma proc_own_contains γp h p v π :
    own γp (ProcAuth h) ∗ own γp (Proc p π v) ⊢ 
     ⌜(∃ π', h !! p = Some (π',DecAgree v))⌝.
  Proof.
    rewrite /ProcAuth /Proc.
    rewrite -own_op.
    iIntros "H".
    iDestruct (own_valid with "H") as %[H1 H2]%auth_valid_discrete_2.
    apply singleton_included in H1 as [n [H4%leibniz_equiv H5]].
    apply Some_included in H5 as [?%leibniz_equiv|?].
    + subst. eauto.
    + specialize (H2 p); rewrite H4 in H2. destruct n as [? []]; last by inversion H2.
      apply prod_included in H as [_ Hp2%DecAgree_included]; subst.
      iPureIntro; eauto.
  Qed.      

  Lemma proc_own_update_ghost γp h p v v' :
    own γp (ProcAuth h) ∗ own γp (Proc p 1 v) ⊢
    |==> own γp (ProcAuth (<[ p := (1%Qp, DecAgree v') ]> h)) ∗ 
    own γp (Proc p 1 v').
  Proof.
    rewrite /ProcAuth /Proc -own_op -own_op.
    iIntros "H".
    iDestruct (own_valid with "H") as %[H1 H2]. simpl in H1.
    rewrite left_id_L in H1.
    specialize (H1 O).
    apply singleton_includedN in H1 as [n [H4 H5]].
    assert (Hxvalid := lookup_valid_Some h p n H2 H4). 
    apply timeless, leibniz_equiv in H4; last typeclasses eauto.
    destruct H5.
    apply Some_timeless in H; last typeclasses eauto.
    apply Some_valid in Hxvalid.
    apply leibniz_equiv in H.
    rewrite H in Hxvalid.
    apply exclusive_Some_l in Hxvalid; last typeclasses eauto. subst.
    inversion H; subst.
    iApply (own_update with "H").
    apply auth_update.
    apply (singleton_local_update _ _ (1%Qp, DecAgree v)); trivial.
    apply exclusive_local_update.
    split; done.
  Qed.

  Lemma proc_too_many_fractions γp p v v' :
    own γp (Proc p ¾ v) ∗ own γp (Proc p ¾ v') 
        ⊢ False.
  Proof.
    iIntros "(H0 & H1)".
    iDestruct (proc_own_combine with "[$H0 $H1]") as "(% & H1)". 
    iDestruct (own_valid with "H1") as %Hvalid. 
      by apply singleton_valid in Hvalid as [[] ?]. 
  Qed.
  
End monoids.
